/** \file
 *  \brief Main form module
 */

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vcl.h>
#pragma hdrstop

#include "Main.h"
#include "display\DisplayOpenGL.h"
#include "display\ValueStringFormatter.h"
#include "FFT.h"
#include "Log.h"
#include "LogUnit.h"
#include "ScopedLock.h"
#include "Mutex.h"
#include "AboutUnit.h"
#include "MyTrackBar.h"

#include "device\MiniscopeDeviceInterface.h"

#include "SettingsUnit.h"
#include "Settings.h"

#include "StatisticsUnit.h"

#include <SysUtils.hpp>
#include <ClipBrd.hpp>
#include <math.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
Settings appSettings;   ///< application settings instance

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner), fZoomChangeGrid(1.25f), device(NULL), bPaintUpdate(false)
{
	CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	CLog::Instance()->SetLevel(E_LOG_TRACE);

	display = new DisplayOpenGL(pnlCenter->Handle);
	// some random settings
	const int iRange = 128;
	const float dy = 3.34e-4f;
	display->SetDiv(0.01f, dy);	// FS = 100Hz
	display->SetRange(iRange);	// 8bit -> +-128
    display->data = &data;
	btnModeScope->Down = true;

#ifdef CREATE_TEST_DATA
	// create and set some data
	data.clear();
	display->data = &data;
	for (int i=0; i<1024; i++)
	{
		data.push_back(	(float)iRange/2 * (	sin(2.0f * M_PI * (float)i/10.24f)           // ~0.1 FS (no leak),
											+ sin(2.0f * M_PI * (float)i/40.0f)         // 0.025 FS
											+ sin(2.0f * M_PI * (float)i/5.0f))+0.5f);	// 0.2 FS
	}
	// FFT
	float *ReIn = new float[data.size()];
	float *ReOut = new float[data.size()];
	float *ImOut = new float[data.size()];
	for(unsigned int i=0;i < data.size();i++)
		ReIn[i]= (float)data[i];
	FFT::Transform(data.size(), ReIn, NULL, ReOut, ImOut);
	dataFFT.resize(data.size()/2);
	//LOG(E_LOG_TRACE, "started");
	for(unsigned int i=0;i < data.size()/2;i++)
	{
		dataFFT[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]));
		//LOG(E_LOG_TRACE, "%f", dataFFT[i]);
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
#endif

	display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

	display->SetMode(Display::MODE_OSCILLOSCOPE);		
	iStoredTriggerTypePos = -1;	

	// load custom cursors
	Screen->Cursors[CURSOR_HAND] = LoadCursor(HInstance, "HANDCUR");
	Screen->Cursors[CURSOR_ZOOM] = LoadCursor(HInstance, "ZOOMCUR");

	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
	btnMouseCursor->Down = true;

	bMovingPlot = false;

	displayOsc.fZoomX = 1.0;
	displayOsc.fZoomY = 1.0;
	displaySpe.fZoomX = 1.0;
	displaySpe.fZoomY = 1.0;
	UpdateZoomDisplay();

	float fValX, fValY;
	display->SetCursor1(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 1, fValX, fValY);
	display->SetCursor2(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 2, fValX, fValY);

	Timer->Interval = 50;
}
//---------------------------------------------------------------------------

__fastcall TfrmMain::~TfrmMain()
{
	CLog::Instance()->Destroy();
}

void __fastcall TfrmMain::pnlCenterResize(TObject *Sender)
{
	if (display)
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::TimerTimer(TObject *Sender)
{
	// timer seems to be better place for any error prone initialization
	// than constructor or OnShow event
	Timer->Enabled = false;
	static bool once = false;
	if (!once)
	{
		once = true;
		// frmLog is already created - set callback to preview log entries
		CLog::Instance()->callbackLog = frmLog->OnLog;
		frmLog->callbackClose = CallbackCloseLogFn;

		display->SetVisibility(Display::MODE_OSCILLOSCOPE, appSettings.GraphOscilloscope.bGridX,
			appSettings.GraphOscilloscope.bGridY, appSettings.GraphOscilloscope.bGridLabels);
		display->SetVisibility(Display::MODE_SPECTRUM, appSettings.GraphSpectrum.bGridX,
			appSettings.GraphSpectrum.bGridY, appSettings.GraphSpectrum.bGridLabels);

		// load device dlls
		asDevDllDir = ExtractFileDir(Application->ExeName) + "\\device";
		DeviceInterface::EnumerateDlls(asDevDllDir);

		Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

		frmStatistics->Left = appSettings.Statistics.iPosX;
		frmStatistics->Top = appSettings.Statistics.iPosY;
		if (appSettings.Statistics.bVisible)
			frmStatistics->Show();

		if (appSettings.asInputDeviceDllName != "")
		{
			AnsiString dllname = asDevDllDir + "\\" + appSettings.asInputDeviceDllName;
			for (unsigned int i=0; i<DeviceInterface::dlls.size(); i++)
			{
				if (DeviceInterface::dlls[i].name == dllname)
				{
					LoadDevice(dllname);
					break;
				}
			}
		} else {
			FirstRunDialog();
		}
		SetMode(Display::MODE_OSCILLOSCOPE);		
	}

	{
		ScopedLock<Mutex> lock(mutexDeviceData);
		if (bDataUpdate)
		{
			unsigned int iExpectedFrameSize = device->capabilities->iBufferSizes[cbBufferSize->ItemIndex];
			if (display->GetMode() == Display::MODE_SPECTRUM)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(data);
				CalculateSpectrum();
			}
			else if (display->GetMode() == Display::MODE_RECORDER)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(dataRec);
				if (data.size() == iExpectedFrameSize)
				{
					std::vector<float>::const_iterator it;
					//float fDivX = device->capabilities->fSensitivityLevels[trbarSensitivity->Position];
					float fAvg = 0;
					for (it = data.begin(); it != data.end(); ++it)
					{
						fAvg += (*it);// * fDivX;
					}
					fAvg /= data.size();
					dataRec.push_back(fAvg);
				}
			}
			else if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(data);
			}
			ProgressBarFrameRx->Position = 100*data.size()/iExpectedFrameSize;
			ProgressBarFrameRx->Update();
		}
		if (bDataUpdate || bPaintUpdate)
		{
			display->Update(Display::UPDATE_ALL);
			bPaintUpdate = false;
			bDataUpdate = false;
		}
	}
	Timer->Enabled = true;
}
//---------------------------------------------------------------------------

void TfrmMain::SetCapabilities(const struct S_SCOPE_CAPABILITIES* const capabilities)
{
	//capabilities->iChannelCount
	display->SetRange(pow(2, capabilities->iBitsPerSample - 1));
	if (capabilities->iSensitivityLevelsCount > 1) {
		lblSensitivity->Visible = true;
		lblSensitivityVal->Visible = true;
		trbarSensitivity->Visible = true;
		/** \todo etykieta */
	} else {
		lblSensitivity->Visible = false;
		lblSensitivityVal->Visible = false;
		trbarSensitivity->Visible = false;
	}
	trbarSensitivity->Max = capabilities->iSensitivityLevelsCount - 1;
	trbarSensitivity->Position = trbarSensitivity->Max;

	if (capabilities->iSamplingPeriodCount > 1) {
		lblSampling->Visible = true;
		lblSamplingVal->Visible = true;
		trbarSampling->Visible = true;
	} else {
		lblSampling->Visible = false;
		lblSamplingVal->Visible = false;
		trbarSampling->Visible = false;
	}
	trbarSampling->Max = capabilities->iSamplingPeriodCount - 1;
	trbarSampling->Position = trbarSampling->Max;

	//trbarTriggerLevel->Min = pow(2, capabilities->iBitsPerSample)/(-2);
	//trbarTriggerLevel->Max = pow(2, capabilities->iBitsPerSample)/2 - 1;
	trbarTriggerLevel->Min = -100;
	trbarTriggerLevel->Max = 100;
	trbarTriggerLevel->Position = 0;

	this->cbTriggerType->Clear();
	ActionModeRecorder->Enabled = false;
	for (unsigned int i=0; i<capabilities->iTriggerTypeCount; i++) {
		switch (capabilities->iTriggerTypes[i]) {
		case TRIGGER_MANUAL:
			cbTriggerType->Items->Add("Manual");
			ActionModeRecorder->Enabled = true;
			break;
		case TRIGGER_SLOPE_UP:
			cbTriggerType->Items->Add("Raising slope");
			break;
		case TRIGGER_SLOPE_DOWN:
			cbTriggerType->Items->Add("Falling slope");
			break;
		case TRIGGER_CONTINUOUS:
			cbTriggerType->Items->Add("Continuous");
			break;
		default:
			break;
		}
	}
	if (capabilities->iTriggerTypeCount > 0) {
		cbTriggerType->ItemIndex = 0;
	}

	this->cbCouplingType->Clear();
	for (unsigned int i=0; i<capabilities->iCouplingTypeCount; i++) {
		switch (capabilities->iCouplingType[i]) {
		case COUPLING_GND:
			cbCouplingType->Items->Add("GND");
			break;
		case COUPLING_DC:
			cbCouplingType->Items->Add("DC");
			break;
		case COUPLING_AC:
			cbCouplingType->Items->Add("AC");
			break;
		default:
			break;
		}
	}
	if (capabilities->iCouplingTypeCount > 0) {
		cbCouplingType->ItemIndex = 0;
	}

	this->cbBufferSize->Clear();
	for (unsigned int i=0; i<capabilities->iBufferSizeCount; i++) {
		cbBufferSize->Items->Add(capabilities->iBufferSizes[i]);
	}
	if (capabilities->iBufferSizeCount > 0) {
    	cbBufferSize->ItemIndex = 0;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::CalculateSpectrum(void)
{
	// FFT
	float *ReIn = new float[data.size()];
	float *ReOut = new float[data.size()];
	float *ImOut = new float[data.size()];
	for(unsigned int i=0;i < data.size();i++)
		ReIn[i]= (float)data[i];
	FFT::Transform(data.size(), ReIn, NULL, ReOut, ImOut);
	dataFFT.resize(data.size()/2);
	//LOG(E_LOG_TRACE, "started");
	for(unsigned int i=0;i < data.size()/2;i++)
	{
		dataFFT[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]));
		//LOG(E_LOG_TRACE, "%f", dataFFT[i]);
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
}
//---------------------------------------------------------------------------



void __fastcall TfrmMain::pnlCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (mouse_tool == MOUSE_CURSOR )
	{
		float fValX, fValY;
		if(Button==mbLeft)  //left button clicked
		{
			display->SetCursor1(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 1, fValX, fValY);
		}
		else if(Button==mbRight)
		{
			display->SetCursor2(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 2, fValX, fValY);
		}
	}
	else if (mouse_tool == MOUSE_OFFSET)
	{
		iPlotMoveStartX = X;
		iPlotMoveStartY = Y;
		bMovingPlot = true;
	}
	else if (mouse_tool == MOUSE_ZOOM || mouse_tool == MOUSE_ZOOM_X || mouse_tool == MOUSE_ZOOM_Y)
	{
		if(Button==mbLeft)  //left button clicked
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX *= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY *= fZoomChangeGrid;
		}
		else if(Button==mbRight)
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX /= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY /= fZoomChangeGrid;
		}
		display->SetZoom(X, Y, DisplayState().fZoomX, DisplayState().fZoomY);
		UpdateZoomDisplay();
	}
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::pnlCenterMouseUp(TObject *Sender, TMouseButton Button,
	  TShiftState Shift, int X, int Y)
{
	if (bMovingPlot)
	{
		display->Move(iPlotMoveStartX, iPlotMoveStartY, X, Y);
		bMovingPlot = false;
		bPaintUpdate = true;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateCursorReadings(enum Display::E_DISPLAY_MODE mode, int cursorId, float fValX, float fValY)
{
	static float fValX1 = 0.0, fValY1 = 0.0, fValX2 = 0.0, fValY2 = 0.0;

	if (mode==Display::MODE_OSCILLOSCOPE || mode==Display::MODE_RECORDER)
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::TimeToStr(dx, 3) + " dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
		if (dx != 0)
		{
			StatusBar->Panels->Items[2]->Text = StatusBar->Panels->Items[2]->Text +
				" 1/dx=" + ValueStringFormatter::FreqToStr(1/dx, 3).c_str();
		}
	}
	else
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::FreqToStr(dx, 3) + "  dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
	}
}


void __fastcall TfrmMain::ActionChangeMouseModeCursorExecute(TObject *Sender)
{
	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeOffsetExecute(TObject *Sender)
{
	mouse_tool = MOUSE_OFFSET;
	pnlCenter->Cursor = (TCursor)CURSOR_HAND;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomXExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_X;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomYExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_Y;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionZoomClearExecute(TObject *Sender)
{
	if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
	{
		displayOsc.fZoomX = 1.0;
		displayOsc.fZoomY = 1.0;
		display->SetZoom(-1, -1, displayOsc.fZoomX, displayOsc.fZoomY);
	}
	else
	{
		displaySpe.fZoomX = 1.0;
		displaySpe.fZoomY = 1.0;
		display->SetZoom(-1, -1, displaySpe.fZoomX, displaySpe.fZoomY);
	}
	UpdateZoomDisplay();
	bPaintUpdate = true;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeScopeExecute(TObject *Sender)
{
	SetMode(Display::MODE_OSCILLOSCOPE);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeSpectrumExecute(TObject *Sender)
{
	SetMode(Display::MODE_SPECTRUM);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeRecorderExecute(TObject *Sender)
{
	SetMode(Display::MODE_RECORDER);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAboutClick(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------



void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
    AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.Read(asConfigFile);
	this->Width = appSettings.iWidth;
	this->Height = appSettings.iHeight;
	this->Top = appSettings.iPosY;
	this->Left = appSettings.iPosX;
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.bWindowMaximized)
		this->WindowState = wsMaximized;
	edInterval->Text = AnsiString::FloatToStrF(appSettings.fRecorderInterval, AnsiString::sffGeneral, 1, 3);
	((TMyTrackBar*)trbarSensitivity)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarSampling)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarTriggerLevel)->OnMouseUp = TrackBarMouseUp;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miDeviceSettingsClick(TObject *Sender)
{
	if (device)
		device->ShowSettings();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miViewLogClick(TObject *Sender)
{
	frmLog->Visible = miViewLog->Checked;
	if (frmLog->Visible)
		frmLog->BringToFront();
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackCloseLogFn(void)
{
	miViewLog->Checked = false;
}

void TfrmMain::CallbackDataFn(short* buf, int size)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	for (unsigned int i=0; i<size/sizeof(short); i++)
	{
		data.push_back(buf[i]);
	}
	bDataUpdate = true;
}

void TfrmMain::CallbackTriggerFn(void)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	data.clear();
	data.reserve(102400);
}

void TfrmMain::CallbackConnectFn(int state, char *szMsgText)
{
	tmrUpdateConnState->Enabled = true;
}

void TfrmMain::UpdateConnectionState(void)
{
	if (!device)
		return;
	struct ConnectionInfo conn = device->GetConnectionInfo();
	switch (conn.state)
	{
	case DEVICE_DISCONNECTED:
		break;
	case DEVICE_CONNECTING:
		break;
	case DEVICE_CONNECTED:
		break;
	case DEVICE_DISCONNECTING:
		break;
	default:
		break;
	}
    StatusBar->Panels->Items[3]->Text = conn.msg;
}

void __fastcall TfrmMain::ActionAcquireStartExecute(TObject *Sender)
{
	if (device)
		device->AcquireStart();
	else
		MessageBox(this->Handle, "No device dll selected.\n"
			"Put dll's in \\device subdir and select with Settings/Common Settings.",
			this->Caption.c_str(), MB_ICONEXCLAMATION);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAcquireStopExecute(TObject *Sender)
{
	if (device)
		device->AcquireStop();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrUpdateConnStateTimer(TObject *Sender)
{
	tmrUpdateConnState->Enabled = false;
	UpdateConnectionState();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbTriggerTypeChange(TObject *Sender)
{
	UpdateTriggerType();
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateTriggerType(void)
{
	if (!device)
		return;
	enum E_TRIGGER_TYPE type = device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex];
	device->SetTriggerType(cbTriggerType->ItemIndex);
	btnManualTrigger->Visible = false;
	trbarTriggerLevel->Visible = false;
	lblTriggerLevel->Visible = false;
	lblTriggerLevelVal->Visible = false;
	btnStart->Visible = false;
	btnStop->Visible = false;
	switch (type) {
	case TRIGGER_MANUAL:
		btnManualTrigger->Visible = true;
		break;
	case TRIGGER_SLOPE_UP:
		trbarTriggerLevel->Visible = true;
		lblTriggerLevel->Visible = true;
		lblTriggerLevelVal->Visible = true;
		break;
	case TRIGGER_SLOPE_DOWN:
		trbarTriggerLevel->Visible = true;
		lblTriggerLevel->Visible = true;
		lblTriggerLevelVal->Visible = true;
		break;
	case TRIGGER_CONTINUOUS:
		btnStart->Visible = true;
		btnStop->Visible = true;
		break;
	default:
		break;
	}
}

void TfrmMain::UpdateCouplingType(void)
{
	if (!device)
		return;
	device->SetCouplingType(cbCouplingType->ItemIndex);
}

void __fastcall TfrmMain::btnManualTriggerClick(TObject *Sender)
{
	if (!device)
		return;
	device->ManualTrigger();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbBufferSizeChange(TObject *Sender)
{
	UpdateBufferSize();
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateBufferSize(void)
{
	if (!device || !cbBufferSize->Items->Count)
		return;
	device->SetBufferSize(cbBufferSize->ItemIndex);
}

void TfrmMain::UpdateSensitivity(void)
{
	if (!device)
		return;
	device->SetSensitivity(trbarSensitivity->Position);
	// update range caption, lblSensitivityVal
	float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
		device->capabilities->fSensitivityLevels[trbarSensitivity->Position];
	lblSensitivityVal->Caption =
		ValueStringFormatter::VoltageToStr(fDivVal, 1).c_str();
	lblSensitivityVal->Caption = lblSensitivityVal->Caption + "/div";

	if (display)
	{
		display->SetDiv(device->capabilities->fSamplingPeriod[trbarSampling->Position],
			device->capabilities->fSensitivityLevels[trbarSensitivity->Position]);
	}
	DataSink::fDivX = device->capabilities->fSensitivityLevels[trbarSensitivity->Position];
}

void TfrmMain::UpdateSampling(void)
{
	if (!device)
		return;
	device->SetSampling(trbarSampling->Position);
	float fSamplVal = (float)1.0 / device->capabilities->fSamplingPeriod[trbarSampling->Position];
	lblSamplingVal->Caption =
		ValueStringFormatter::FreqToStr(fSamplVal, 1).c_str();

	if (display)
	{
		display->SetDiv(device->capabilities->fSamplingPeriod[trbarSampling->Position],
			device->capabilities->fSensitivityLevels[trbarSensitivity->Position]);
	}
}

void TfrmMain::UpdateTriggerLevel(void)
{
	if (!device)
		return;
#if 0
	device->SetTriggerLevel(trbarTriggerLevel->Position);
	int val;
	if (trbarTriggerLevel->Position >= 0)
		val = 100 * (float)trbarTriggerLevel->Position/trbarTriggerLevel->Max;
	else
		val = -100 * (float)trbarTriggerLevel->Position/trbarTriggerLevel->Min;
	lblTriggerLevelVal->Caption = IntToStr(val) + "%";
#else
	device->SetTriggerLevel(trbarTriggerLevel->Position);
	lblTriggerLevelVal->Caption = IntToStr(trbarTriggerLevel->Position) + "%";
#endif
}

void __fastcall TfrmMain::cbCouplingTypeChange(TObject *Sender)
{
	UpdateCouplingType();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterPaint(TObject *Sender)
{
	/** \note PaintBoxCenter only purpose is to fire OnPaint event
		to request graph repaint (TPanel has no OnPaint event but
		TPaintBox has no Handle
    */
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseDown(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseUp(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseUp(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	delete display;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}

	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	//appSettings.bSendOnEnter = btnSendOnEnter->Down;
	appSettings.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.iWidth = this->Width;
		appSettings.iHeight = this->Height;
		appSettings.iPosY = this->Top;
		appSettings.iPosX = this->Left;
	}

	appSettings.Statistics.iPosX = frmStatistics->Left;
	appSettings.Statistics.iPosY = frmStatistics->Top;
	appSettings.Statistics.bVisible = frmStatistics->Visible;

	appSettings.fRecorderInterval = StrToFloatDef(edInterval->Text, 10.0);

	appSettings.Write(asConfigFile);

	CanClose = true;
}
//---------------------------------------------------------------------------

void TfrmMain::LoadDevice(AnsiString dllname)
{
	ToolBarMode->Enabled = false;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}
	data.clear();
	dataFFT.clear();
	iStoredTriggerTypePos = -1;
	device = new DeviceInterface(dllname);
	if (device->Load())
	{
		ShowMessage("Failed loading device dll.");
		delete device;
		device = NULL;
		return;
	}
	device->callbackData = CallbackDataFn;
	device->callbackConnect = CallbackConnectFn;
	device->callbackTrigger = CallbackTriggerFn;

	display->SetDiv(device->capabilities->fSamplingPeriod[0],
					device->capabilities->fSensitivityLevels[0]);
	SetCapabilities(device->capabilities);

	if (device->GetSettings(&device->settings) == 0)
	{
		trbarSensitivity->Position = device->settings.iSensitivity;
		trbarSampling->Position = device->settings.iSamplingPeriod;
		trbarTriggerLevel->Position = device->settings.iTriggerLevel;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == device->settings.eTriggerType)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		for (unsigned int i=0; i<device->capabilities->iCouplingTypeCount; i++)
		{
			if (device->capabilities->iCouplingType[i] == device->settings.eCouplingType)
			{
				cbCouplingType->ItemIndex = i;
				break;
			}
		}
		cbBufferSize->ItemIndex = cbBufferSize->Items->IndexOf(device->settings.iBufferSize);
	}
	if (device->Connect())
	{
		ShowMessage("Failed connecting to device.");
	} else
	{
		UpdateTriggerType();
		UpdateCouplingType();
		UpdateBufferSize();
		UpdateSensitivity();
		UpdateSampling();
		UpdateTriggerLevel();
		ToolBarMode->Enabled = true;
	}
}

std::auto_ptr<TImage> TfrmMain::CreatePlotImage(void)
{
	bool bStatState = frmStatistics->Visible;
	if (bStatState)
	{
        // hide statistics window
		frmStatistics->Visible = false;
		display->Update(Display::UPDATE_ALL);
	}
	frmStatistics->Update();
	std::auto_ptr<TImage> Image(new TImage(NULL));
	HDC SrcDC, DestDC;
	Image->Width = pnlCenter->Width;
	Image->Height = pnlCenter->Height;
	SrcDC = GetDC(pnlCenter->Handle);
	DestDC = Image->Canvas->Handle;
	BitBlt(DestDC,
		  0, 0,
		  Image->Width, Image->Height,
		  SrcDC,
		  0, 0,
		  SRCCOPY);
	ReleaseDC(pnlCenter->Handle, SrcDC);
	if (bStatState)
	{
		frmStatistics->Visible = bStatState;
		this->SetFocus();
	}
	return Image;
}

void __fastcall TfrmMain::ActionCopyPlotClipbrdExecute(TObject *Sender)
{
	std::auto_ptr<TImage> Image = CreatePlotImage();
	TClipboard *pCB = Clipboard();
	pCB->Assign(Image->Picture);
	(void)pCB;	// avoid unused variable warning
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateZoomDisplay(void)
{
	lblZoomX->Caption = (AnsiString)"X: " + (ValueStringFormatter::ValToStr(DisplayState().fZoomX, 2) + "x").c_str();
	lblZoomY->Caption = (AnsiString)"Y: " + (ValueStringFormatter::ValToStr(DisplayState().fZoomY, 2) + "x").c_str();
}


void __fastcall TfrmMain::TrackbarKeyUp(TObject *Sender, WORD &Key,
	  TShiftState Shift)
{
	if (!device)
		return;
   	UpdateTrackBar(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::TrackBarMouseUp(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (!device)
		return;
	UpdateTrackBar(Sender);
}

void TfrmMain::UpdateTrackBar(TObject *Sender)
{
	if (Sender == trbarSensitivity)
	{
		UpdateSensitivity();
	}
	else if (Sender == trbarSampling)
	{
		UpdateSampling();
	}
	else if (Sender == trbarTriggerLevel)
	{
		UpdateTriggerLevel();
	}
}

void __fastcall TfrmMain::miStatisticsClick(TObject *Sender)
{
	frmStatistics->Show();	
}
//---------------------------------------------------------------------------

void TfrmMain::FirstRunDialog(void)
{
	const char msg[] = "No input device dll is selected.\n"
		"It seems like application is running for ther first time\n"
		"or configuration has been removed.\n"
		"Would you like to select input device dll now?";
	if (MessageBox(this->Handle, msg,
		"No input device plugin selected", MB_YESNO | MB_ICONQUESTION) == ID_YES)
	{
	    ActionShowCommonSettingsExecute(NULL);
	}
}

void __fastcall TfrmMain::ActionShowCommonSettingsExecute(TObject *Sender)
{
	AnsiString asPrevDevice = appSettings.asInputDeviceDllName;
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.asInputDeviceDllName != asPrevDevice)
		LoadDevice(asDevDllDir + "\\" + appSettings.asInputDeviceDllName);
	Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
	
	display->SetVisibility(Display::MODE_OSCILLOSCOPE, appSettings.GraphOscilloscope.bGridX,
		appSettings.GraphOscilloscope.bGridY, appSettings.GraphOscilloscope.bGridLabels);
	display->SetVisibility(Display::MODE_SPECTRUM, appSettings.GraphSpectrum.bGridX,
		appSettings.GraphSpectrum.bGridY, appSettings.GraphSpectrum.bGridLabels);
	bPaintUpdate = true;	
}
//---------------------------------------------------------------------------

void TfrmMain::SetMode(enum Display::E_DISPLAY_MODE mode)
{
	int oldmode = -1;

    // restore trigger type to state before switching to recorder mode
	if (mode != oldmode && oldmode == Display::MODE_RECORDER)
	{
		if (cbTriggerType->Items->Count > iStoredTriggerTypePos && iStoredTriggerTypePos != -1)
		{
			cbTriggerType->ItemIndex = iStoredTriggerTypePos;		
			UpdateTriggerType();
		}
	}

	bDataUpdate = true;
	if (mode == Display::MODE_OSCILLOSCOPE)
		display->data = &data;
	else if (mode == Display::MODE_SPECTRUM)
		display->data = &dataFFT;
	else if (mode == Display::MODE_RECORDER)
	{
		display->data = &dataRec;
		iStoredTriggerTypePos = cbTriggerType->ItemIndex;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == TRIGGER_MANUAL)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		UpdateTriggerType();
	}
	display->SetMode(mode);
	UpdateZoomDisplay();


	if (mode != Display::MODE_RECORDER)
	{
		cbTriggerType->Enabled = true;
		if (device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex] == TRIGGER_MANUAL)
			btnManualTrigger->Visible = true;
		btnRecorderStart->Visible = false;
		btnRecorderStop->Visible = false;
		btnRecorderClear->Visible = false;
		lblInterval->Visible = false;
		edInterval->Visible = false;
		lblSeconds->Visible = false;

		tmrRecorder->Enabled = false;
		btnRecorderStart->Enabled = true;
		btnRecorderStop->Enabled = false;
	}
	else
	{
		cbTriggerType->Enabled = false;
		btnManualTrigger->Visible = false;
		btnRecorderStart->Visible = true;
		btnRecorderStop->Visible = true;
		btnRecorderClear->Visible = true;
		lblInterval->Visible = true;
		edInterval->Visible = true;
		lblSeconds->Visible = true;
	}	
}


void __fastcall TfrmMain::edIntervalChange(TObject *Sender)
{
	bool bPrevState = tmrRecorder->Enabled;
	tmrRecorder->Enabled = false;
	tmrRecorder->Interval = StrToFloatDef(edInterval->Text, 10.0) * 1000.0;
	tmrRecorder->Enabled = bPrevState;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderStartClick(TObject *Sender)
{
	tmrRecorder->Enabled = true;
	btnRecorderStart->Enabled = false;
	btnRecorderStop->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderStopClick(TObject *Sender)
{
	tmrRecorder->Enabled = false;
	btnRecorderStart->Enabled = true;
	btnRecorderStop->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderClearClick(TObject *Sender)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	dataRec.clear();
	bDataUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrRecorderTimer(TObject *Sender)
{
	assert(display->GetMode() == Display::MODE_RECORDER);
	// execute manual trigger
	if (device)
		device->ManualTrigger();
	/** \todo check if previous sample already received */
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionExportPlotBitmapExecute(TObject *Sender)
{
	SaveDialog->DefaultExt = "bmp";
	SaveDialog->Filter = "Bitmap (*.bmp)|*.bmp";
	SaveDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotPictureExportDir))
		SaveDialog->InitialDir = appSettings.asPlotPictureExportDir;
	SaveDialog->FileName = SaveDialog->InitialDir + "\\" + FormatDateTime("yyyymmdd_hhnn", Now());
	if (SaveDialog->Execute())
	{
		std::auto_ptr<TImage> Image = CreatePlotImage();
		Image->Picture->SaveToFile(SaveDialog->FileName);
		appSettings.asPlotPictureExportDir = ExtractFileDir(SaveDialog->FileName);
	}
}
//---------------------------------------------------------------------------

